from pydantic import BaseModel


class NearTrustTransaction(BaseModel):
    from_account_id: str
    amount: str


class FtTrustTransaction(BaseModel):
    from_account_id: str
    ft_contract_id: str
    ft_amount: str


class NftTrustTransaction(BaseModel):
    from_account_id: str
    nft_contract_id: str
    nft_token_id: str



from pynear.dapps.ft.exceptions import NotRegisteredError, NotEnoughBalance
from pynear.exceptions.exceptions import FunctionCallError

from pynear.dapps.core import DappClient, NEAR
from pynear.dapps.ft.models import FtTokenMetadata
from pynear.dapps.fts import FtModel


class FT(DappClient):
    async def get_ft_balance(self, ft: FtModel, account_id: str) -> float:
        return (
            await self._account.view_function(
                ft.contract_id,
                "ft_balance_of",
                {"account_id": account_id},
            )
        ).result // 10**ft.decimal

    async def get_metadata(self, ft: FtModel) -> FtTokenMetadata:
        return FtTokenMetadata(
            **(
                await self._account.view_function(
                    ft.contract_id,
                    "ft_metadata",
                    {},
                )
            ).result
        )

    async def transfer(self, ft: FtModel, receiver_id: str, amount: float, memo: str = ""):
        try:
            return await self._account.function_call(
                ft.contract_id,
                "ft_transfer",
                {
                    "receiver_id": receiver_id,
                    "amount": str(int(amount * 10**ft.decimal)),
                    "msg": memo,
                },
                amount=1,
            )
        except FunctionCallError as e:
            if "The account is not registered" in e.error["ExecutionError"]:
                raise NotRegisteredError(e)
            if "The account doesn't have enough balance" in e.error["ExecutionError"]:
                raise NotEnoughBalance(e)
            raise e

    async def transfer_call(self, ft: FtModel, receiver_id: str, amount: float, memo: str = ""):
        return await self._account.function_call(
            ft.contract_id,
            "ft_transfer_call",
            {
                "receiver_id": receiver_id,
                "amount": str(int(amount * 10**ft.decimal)),
                "msg": memo,
            },
            amount=1,
        )

    async def storage_balance_of(self, contract_id, near_account_id: str):
        return (
            await self._account.view_function(
                contract_id,
                "storage_balance_of",
                {"account_id": near_account_id},
            )
        ).result

    async def storage_deposit(self, ft: FtModel, near_account_id: str, amount: int = NEAR // 100):
        return await self._account.function_call(
            ft.contract_id,
            "storage_deposit",
            {"account_id": near_account_id},
            amount=amount,
        )



import json

import base58
import ed25519


class KeyPair(object):
    def __init__(self, secret_key):
        secret_key = secret_key.split(":")[1] if ":" in secret_key else secret_key
        self._secret_key = ed25519.SigningKey(base58.b58decode(secret_key))
        self._public_key = self._secret_key.get_verifying_key()

    @property
    def public_key(self):
        return self._public_key.to_bytes()

    def encoded_public_key(self):
        return base58.b58encode(self._public_key.to_bytes()).decode("utf-8")

    def sign(self, message):
        return self._secret_key.sign(message)


class Signer(object):
    def __init__(self, account_id, key_pair):
        self._account_id = account_id
        self._key_pair = key_pair

    @property
    def account_id(self):
        return self._account_id

    @property
    def key_pair(self):
        return self._key_pair

    @property
    def public_key(self):
        return self._key_pair.public_key

    def sign(self, message):
        return self._key_pair.sign(message)

    @classmethod
    def from_json(self, j):
        return Signer(j["account_id"], KeyPair(j["secret_key"]))

    @classmethod
    def from_json_file(self, json_file):
        with open(json_file) as f:
            return Signer.from_json(json.loads(f.read()))



class BinarySerializer:
    def __init__(self, schema):
        self.array = bytearray()
        self.schema = schema

    def serialize_num(self, value, n_bytes):
        orig_value = value
        assert value >= 0, "Can't serialize negative numbers %d" % value
        for i in range(n_bytes):
            self.array.append(value & 255)
            value //= 256
        assert value == 0, "Value %d has more than %d bytes" % (orig_value, n_bytes)

    def serialize_field(self, value, fieldType):
        try:
            if type(fieldType) == str:
                if fieldType[0] == "u":
                    self.serialize_num(value, int(fieldType[1:]) // 8)
                elif fieldType == "string":
                    b = value.encode("utf8")
                    self.serialize_num(len(b), 4)
                    self.array += b
                else:
                    assert False, fieldType
            elif type(fieldType) == list:
                assert len(fieldType) == 1
                if type(fieldType[0]) == int:
                    assert type(value) == bytes, "type(%s) = %s != bytes" % (
                        value,
                        type(value),
                    )
                    assert len(value) == fieldType[0], "len(%s) = %s != %s" % (
                        value,
                        len(value),
                        fieldType[0],
                    )
                    self.array += bytearray(value)
                else:
                    self.serialize_num(len(value), 4)
                    for el in value:
                        self.serialize_field(el, fieldType[0])
            elif type(fieldType) == dict:
                assert fieldType["kind"] == "option"
                if value is None:
                    self.serialize_num(0, 1)
                else:
                    self.serialize_num(1, 1)
                    self.serialize_field(value, fieldType["type"])
            elif type(fieldType) == type:
                assert type(value) == fieldType, "%s != type(%s)" % (fieldType, value)
                self.serialize_struct(value)
            else:
                assert False, type(fieldType)
        except:
            print("Failed to serialize %s as %s" % (value, fieldType))
            raise

    def serialize_struct(self, obj):
        structSchema = self.schema[type(obj)]
        if structSchema["kind"] == "struct":
            for fieldName, fieldType in structSchema["fields"]:
                self.serialize_field(getattr(obj, fieldName), fieldType)
        elif structSchema["kind"] == "enum":
            name = getattr(obj, structSchema["field"])
            for idx, (fieldName, fieldType) in enumerate(structSchema["values"]):
                if fieldName == name:
                    self.serialize_num(idx, 1)
                    self.serialize_field(obj.data, fieldType)
                    break
        else:
            assert False, structSchema

    def serialize(self, obj):
        self.serialize_struct(obj)
        return bytes(self.array)



from setuptools import setup
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


extra_files = package_files("pynear")

setup(
    name="py-near",
    author_email="petr@herewallet.app",
    url="https://github.com/pvolnov/py-near",
    packages=["pynear"],
    python_requires=">=3.6",
    package_data={"": extra_files},
    project_urls={
        "Documentation": "https://py-near.readthedocs.io/en/latest",
        "Funding": "https://opencollective.com/py-near",
        "Twitter": "https://twitter.com/p_volnov",
        "Source": "https://github.com/pvolnov/py-near",
        "Tracker": "https://github.com/pvolnov/py-near/issues",
    },
    version="1.0.2",
    description="Pretty simple and fully asynchronous framework for working with NEAR blockchain",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Petr Volnov",
    license="MIT",
    install_requires=["base58", "ed25519", "aiohttp"],
)



import hashlib
from dataclasses import dataclass
from typing import Union, List

import base58

from pynear.serializer import BinarySerializer


@dataclass
class Signature:
    keyType: int
    data: bytes


@dataclass
class PublicKey:
    keyType: int
    data: bytes


@dataclass
class FunctionCallPermission:
    allowance: int
    receiverId: str
    methodNames: List[str]


class FullAccessPermission:
    pass


@dataclass
class AccessKeyPermission:
    enum: str
    data: Union[FunctionCallPermission, FullAccessPermission]


@dataclass
class AccessKey:
    nonce: int
    permission: AccessKeyPermission


class CreateAccount:
    pass


class DeployContract:
    pass


@dataclass
class FunctionCall:
    methodName: str
    args: bytes
    gas: int
    deposit: str


@dataclass
class Transfer:
    deposit: int


@dataclass
class Stake:
    stake: str
    publicKey: PublicKey


@dataclass
class AddKey:
    accessKey: AccessKey
    publicKey: PublicKey


@dataclass
class DeleteKey:
    publicKey: PublicKey


class DeleteAccount:
    pass


@dataclass
class Action:
    enum: str
    data: Union[Transfer, Stake, FunctionCall, AddKey, CreateAccount, DeleteKey, DeployContract]


@dataclass
class Transaction:
    signerId: str
    publicKey: PublicKey
    nonce: int
    receiverId: str
    actions: List[Action]
    blockHash: str


@dataclass
class SignedTransaction:
    transaction: Transaction
    signature: Signature


tx_schema = dict(
    [
        [Signature, {"kind": "struct", "fields": [["keyType", "u8"], ["data", [64]]]}],
        [
            SignedTransaction,
            {
                "kind": "struct",
                "fields": [["transaction", Transaction], ["signature", Signature]],
            },
        ],
        [
            Transaction,
            {
                "kind": "struct",
                "fields": [
                    ["signerId", "string"],
                    ["publicKey", PublicKey],
                    ["nonce", "u64"],
                    ["receiverId", "string"],
                    ["blockHash", [32]],
                    ["actions", [Action]],
                ],
            },
        ],
        [PublicKey, {"kind": "struct", "fields": [["keyType", "u8"], ["data", [32]]]}],
        [
            AccessKey,
            {
                "kind": "struct",
                "fields": [
                    ["nonce", "u64"],
                    ["permission", AccessKeyPermission],
                ],
            },
        ],
        [
            AccessKeyPermission,
            {
                "kind": "enum",
                "field": "enum",
                "values": [
                    ["functionCall", FunctionCallPermission],
                    ["fullAccess", FullAccessPermission],
                ],
            },
        ],
        [
            FunctionCallPermission,
            {
                "kind": "struct",
                "fields": [
                    ["allowance", {"kind": "option", "type": "u128"}],
                    ["receiverId", "string"],
                    ["methodNames", ["string"]],
                ],
            },
        ],
        [FullAccessPermission, {"kind": "struct", "fields": []}],
        [
            Action,
            {
                "kind": "enum",
                "field": "enum",
                "values": [
                    ["createAccount", CreateAccount],
                    ["deployContract", DeployContract],
                    ["functionCall", FunctionCall],
                    ["transfer", Transfer],
                    ["stake", Stake],
                    ["addKey", AddKey],
                    ["deleteKey", DeleteKey],
                    ["deleteAccount", DeleteAccount],
                ],
            },
        ],
        [CreateAccount, {"kind": "struct", "fields": []}],
        [DeployContract, {"kind": "struct", "fields": [["code", ["u8"]]]}],
        [
            FunctionCall,
            {
                "kind": "struct",
                "fields": [
                    ["methodName", "string"],
                    ["args", ["u8"]],
                    ["gas", "u64"],
                    ["deposit", "u128"],
                ],
            },
        ],
        [Transfer, {"kind": "struct", "fields": [["deposit", "u128"]]}],
        [
            Stake,
            {"kind": "struct", "fields": [["stake", "u128"], ["publicKey", PublicKey]]},
        ],
        [
            AddKey,
            {
                "kind": "struct",
                "fields": [["publicKey", PublicKey], ["accessKey", AccessKey]],
            },
        ],
        [DeleteKey, {"kind": "struct", "fields": [["publicKey", PublicKey]]}],
        [DeleteAccount, {"kind": "struct", "fields": [["beneficiaryId", "string"]]}],
    ]
)


def sign_and_serialize_transaction(receiverId, nonce, actions, blockHash, signer):
    if signer.public_key is None:
        raise ValueError("Signer must have a public key")
    if blockHash is None:
        raise ValueError("Block hash is required")
    tx = Transaction(
        signer.account_id,
        PublicKey(0, signer.public_key),
        nonce,
        receiverId,
        actions,
        blockHash,
    )

    msg = BinarySerializer(tx_schema).serialize(tx)
    hash_ = hashlib.sha256(msg).digest()

    signature = Signature(0, signer.sign(hash_))
    signedTx = SignedTransaction(tx, signature)
    return BinarySerializer(tx_schema).serialize(signedTx)


def create_create_account_action():
    createAccount = CreateAccount()
    action = Action("createAccount", createAccount)
    return action


def create_full_access_key_action(pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    permission = AccessKeyPermission("fullAccess", FullAccessPermission())
    accessKey = AccessKey(0, permission)
    publicKey = PublicKey(0, pk)
    addKey = AddKey(accessKey, publicKey)
    action = Action("addKey", addKey)
    return action


def create_function_call_access_key_action(
    pk, allowance: int, receiverId: str, methodNames: List[str]
):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    permission = AccessKeyPermission(
        "functionCall", FunctionCallPermission(allowance, receiverId, methodNames)
    )
    accessKey = AccessKey(0, permission)
    publicKey = PublicKey(0, pk)
    addKey = AddKey(accessKey, publicKey)
    action = Action("addKey", addKey)
    return action


def create_delete_access_key_action(pk):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    publicKey = PublicKey(0, pk)
    deleteKey = DeleteKey(publicKey)
    action = Action("deleteKey", deleteKey)
    return action


def create_transfer_action(amount: int):
    transfer = Transfer(amount)
    action = Action("transfer", transfer)
    return action


create_payment_action = create_transfer_action


def create_staking_action(amount, pk):
    stake = Stake(amount, PublicKey(0, pk))
    action = Action("stake", stake)
    return action


def create_deploy_contract_action(code):
    deployContract = DeployContract()
    deployContract.code = code
    action = Action("deployContract", deployContract)
    return action


def create_function_call_action(method_name, args, gas, deposit):
    function_call = FunctionCall(method_name, args, gas, deposit)
    action = Action("functionCall", function_call)
    return action
